local FRAME_WIDTH = 840
local FRAME_HEIGHT = 550
local PROFILES_CONTENT_WIDTH = 0.73 -- relative to content frame inset
local PROFILES_LISTING_HEIGHT = 24
local PROFILES_LISTING_SPACING = 0
local PROFILES_LISTING_INDENT = 20
local PROFILES_TALENT_WIDTH = 193
local PROFILES_TALENT_HEIGHT = 52
local PROFILES_TALENT_SPACING = 1
local MAX_TREE_DEPTH = 2 -- artificial limit (0 is root)

local DEFAULTS = {
	char = {
		["settings"] = {
			["spec"] = true,
			["talents"] = true,
			["verbose"] = true,
			["debug"] = false
		},
		-- Base profile Structure
		["profiles"] = {
			{
				["type"] = "CATEGORY",
				["name"] = "Arenas",
				["data"] = {
				},
			},
			{
				["type"] = "CATEGORY",
				["name"] = "Battlegrounds",
				["data"] = {	
				},
			},
			{
				["type"] = "CATEGORY",
				["name"] = "Raids",
				["data"] = {
				},
			},
			{
				["type"] = "CATEGORY",
				["name"] = "Dungeons",
				["data"] = {
				},
			},
			{
				["type"] = "CATEGORY",
				["name"] = "Other",
				["data"] = {
				},
			},
		},
	},
}


-- Create Addon Object
BossProfiles = LibStub("AceAddon-3.0"):NewAddon("BossProfiles","AceConsole-3.0")

-- When the addon is initialized
function BossProfiles:OnInitialize()
	-- Register slash commands
	self:RegisterChatCommand("bp","SlashCommand")
	self:RegisterChatCommand("bossprofiles","SlashCommand")

	-- Load settings from db, if no settings, load default
	BossProfiles.db = LibStub("AceDB-3.0"):New("BossProfilesDB", DEFAULTS)
end

function BossProfiles:SlashCommand(input)
	local command, arg = string.split(" ", input, 2)

	-- Get settings ref
	local settings = self.db.char.settings

	if(strlen(command) == 0) then
		if(GetSpecialization(nil,nil,1) ~= nil or GetSpecialization(nil,nil,2) ~= nil) then
			BossProfilesFrame:Show()
		else
			BossProfiles:Print("You must be specced into at least one specialization to configure BossProfiles.")
		end
	elseif(command == "enable") then
		if(settings[arg] ~= nil) then
			self:Print("Enabling: "..arg)
			self.db.char.settings[arg] = true
			return true
		else
			self:Print("\""..arg.."\" is not a valid setting")
			return false
		end
	elseif(command == "disable") then
		if(settings[arg] ~= nil) then
			self:Print("Disabling: "..arg)
			self.db.char.settings[arg] = false
			return true
		else
			self:Print("\""..arg.."\" is not a valid setting, no changes made")
			return false
		end
	elseif(command == "settings") then
		local output = "Current Settings: "
		for setting,state in pairs(settings) do
			if(state) then state = "true" else state = "false" end
			output = output.."\n"..setting..": "..state
		end
		self:Print(output)
	else
		self:Print("BossProfile uses the following syntax: "..
			"\n /bp [command] [argument]"..
			"\n /bp settings"..
			"\n /bp enable settingName"..
			"\n /bp disable settingName")
	end
end

function BossProfiles:OnEvent(self, event, ...)
	if(BossProfilesFrame.initialized) then
		if event == "ACTIVE_TALENT_GROUP_CHANGED" then
			BossProfiles.util:UpdateApplyButton()
		elseif event == "PLAYER_SPECIALIZATION_CHANGED" then
			BossProfiles.role:Update()
			local primarySpec = GetSpecialization(nil,nil,1)
			local secondarySpec = GetSpecialization(nil,nil,2)
			if(primarySpec ~= nil and secondarySpec == nil) then 
				BossProfiles.role:SetSelected(GetSpecializationRole(primarySpec))
			elseif(primarySpec == nil and secondarySpec ~= nil) then 
				BossProfiles.role:SetSelected(GetSpecializationRole(secondarySpec))
			elseif(primarySpec == nil and secondarySpec == nil) then
				BossProfilesFrame:Hide()
			end
			BossProfiles.content.spec:Update()
		elseif event == "PLAYER_TALENT_UPDATE" then
			BossProfiles.content.talents:Update()
			BossProfiles.util:UpdateApplyButton()
		elseif event == "ROLE_CHANGED_INFORM" then
			local arg = {...}
			local name = GetUnitName("player",false)
			if arg[1] == name and arg[2] == name and arg[4] ~= "NONE" then
				BossProfiles.role:SetSelected(arg[4])
			end
		elseif event == "BAG_UPDATE_DELAYED" then
			BossProfiles.util:UpdateReagentCount()
		end
	end
end

function BossProfiles:Print(msg)
	DEFAULT_CHAT_FRAME:AddMessage("|cFF00b2ffBossProfiles: |r"..msg)
end

-- When the addon is enabled
function BossProfiles:OnEnable()
	-- Create frame for applying profile
	CreateFrame("Button","BossProfilesLoader",nil,"SecureActionButtonTemplate")
	CreateFrame("Frame", "BossProfilesMenu")
end

-- When the addon is disabled
function BossProfiles:OnDisable()
  -- Unhook, Unregister Events, Hide frames that were created
end

-- When the interface is shown
function BossProfiles:OnInterfaceShow(self)
	if(not self.initialized) then
		BossProfiles.content:Initialize()
		BossProfiles.profileList:Initialize()
		BossProfiles.content.talents:Initialize()
		BossProfiles.content.spec:Initialize()
		BossProfilesFrame_ProfilesFooterNewCategoryButton:Disable()
		BossProfilesFrame_ProfilesFooterNewProfileButton:Disable()
		BossProfilesFrame_ProfilesFooterApplyButton:SetAttribute("type","macro")
		BossProfilesFrame_ProfilesFooterApplyButton:SetAttribute("macrotext",
			"/stopmacro [combat]\n"..
			"/script BossProfiles.profile:Apply()\n"..
			"/click BossProfilesLoader"
		)
		tinsert(UISpecialFrames, BossProfilesFrame:GetName())

		self.initialized = true
	end

	BossProfiles.role:Update()
	BossProfiles.util.UpdateReagentCount()
	BossProfiles.util:UpdateApplyButton()
end

-- When the addon interface loads
function BossProfiles:OnInterfaceLoad(self)
	-- Initialize frame size
	self:SetSize(FRAME_WIDTH,FRAME_HEIGHT)

	-- Initialize Player class icon
	local class = select(2, UnitClass("player"))
	self.Portrait:SetTexture("Interface\\TargetingFrame\\UI-Classes-Circles")
	self.Portrait:SetTexCoord(unpack(CLASS_ICON_TCOORDS[strupper(class)]))
	
	-- Event Registration
	self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	self:RegisterEvent("ROLE_CHANGED_INFORM")
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
	self:RegisterEvent("BAG_UPDATE_DELAYED")
	self:RegisterForDrag("LeftButton")

	-- Initialize Tabs, todo: re-enable this
	-- PanelTemplates_SetNumTabs(self, 2)
	-- PanelTemplates_SetTab(self, 1)

	BossProfiles.util = {
		-- Takes a "real" spec id as input, returns "relative" spec output:
		-- 1 if primary, 2 if secondary, 0 if inactive
		GetRelativeSpec = function(self,spec)
			if(spec == GetSpecialization(nil,nil,1)) then return 1
			elseif(spec == GetSpecialization(nil,nil,2)) then return 2
			else return 0 end
		end,

		-- Updates the reagent counter
		UpdateReagentCount = function()
			local reagents, icon = select(2,GetTalentClearInfo());
			BossProfilesFrame_ProfilesFooterReagentCount:SetText( format("%d |T%s:13:13:0:0:64:64:5:59:5:59|t",reagents or 0, icon or ''));
		end,

		UpdateApplyButton = function()
			local button = BossProfilesFrame_ProfilesFooterApplyButton
			local selected = BossProfiles.profileList:GetSelected()
			local role = BossProfiles.role:GetSelected()

			-- TODO: update this so that it can loop over classes, if class has "GetChangeCount" function,
			-- check settings to see if enabled and invoke it if so
			if(selected ~= nil and selected.ref.type == "PROFILE" and selected.ref.data[role].activeSpec ~= 0) then
				local profile = selected.ref.data[role]
				local changeCount = 0
				local count, invalidProfile

				if(BossProfiles.db.char.settings.spec) then
					count = BossProfiles.content.spec:GetChangeCount(profile)
					if(count > 0) then
						changeCount = changeCount + count
					elseif(count == -1) then
						invalidProfile = true
					end
				end

				if(BossProfiles.db.char.settings.talents) then
					count = BossProfiles.content.talents:GetChangeCount(profile)
					if(count > 0) then
						changeCount = changeCount + count
					elseif(count == -1) then
						invalidProfile = true
					end
				end

				if(invalidProfile == false) then
					button:SetText("Apply [?]")
					button:Disable()
				else
					button:SetText("Apply ["..changeCount.."]")
					if(changeCount>0) then
						button:Enable()
					else
						button:Disable()
					end
				end
			else
				button:SetText("Apply [0]")
				button:Disable()
			end
		end,

		PreventModification = function()
			local settings = BossProfiles.db.char.settings
			BossProfilesLoader:SetAttribute("type","")
			BossProfilesLoader:SetAttribute("slot","")
			BossProfilesLoader:SetAttribute("glyph","")
			BossProfilesLoader:SetAttribute("macrotext","")
		end,
	}

	BossProfiles.profile = {
		lastLoaded,

		-- Loads the profile for currently select button, intializes it if needed
		Load = function(self)
			local selected = BossProfiles.profileList:GetSelected()
			local role = BossProfiles.role:GetSelected()

			if (role and selected.ref.data[role] ~= nil) then
				BossProfiles.content.spec:SetSelected(selected.ref.data[role].activeSpec, true, "REAL")
			elseif(selected.ref.type == "PROFILE") then
				if(#selected.ref.data == 0) then
					self:Create("SUBPROFILE",profile,role)
					self:Load(profile)
				end
			end

			BossProfiles.content.spec:Update()
		end,

		-- Adds a node to data-structure, placement is relative to currently selected profile. Types:
		-- * CATEGORY: Collapsable heading
		--	* PROFILE: Parent for role subprofiles
		--	* SUBPROFILE: Specific role profile
		Create = function(self,type)
			if(type ~= nil) then
				if(type == "CATEGORY" or type=="PROFILE") then
					local selected = BossProfiles.profileList:GetSelected()
					if(type == "CATEGORY" and selected ~= nil and selected.ref.type~="PROFILE" and selected.depth >= MAX_TREE_DEPTH) then
						print("Unable to create profile, max tree depth reached: "..MAX_TREE_DEPTH)
					else
						-- Determine where to insert the new node
						local node
						if(selected and selected.depth > 0 and selected.ref.type=="PROFILE") then
							node = selected.parent.data
						elseif(not selected or (selected.depth == 0 and selected.ref.type=="PROFILE")) then 
							node = BossProfiles.db.char.profiles
						else 
							node = selected.ref.data
						end

						-- Insert data - after selected element if profile
						if(selected ~= nil and selected.ref.type == "PROFILE") then
							local index
							for k,v in pairs(selected.parent.data) do
								if(v == selected.ref) then
									index = k
									break
								end
							end	
							tinsert(node,index+1,{type=type,data={},renamePending=true})
						else -- if category is selected, insert new profile at end
							tinsert(node,{type=type,data={},renamePending=true})
						end

						-- Actions to perform after adding a category or profile
						BossProfiles.profileList:RefreshData()
						BossProfiles.profileList:Update()
					end
				elseif(type == "SUBPROFILE") then
					local selected = BossProfiles.profileList:GetSelected()
					local role = BossProfiles.role:GetSelected()

					for i=1,GetNumSpecializations() do
						if(GetSpecializationRole(i) == role) then

							if(selected.ref.data[role] == nil) then
								selected.ref.data[role] = {activeSpec = 0}
							end

							selected.ref.data[role][i] = {}
						end
					end
				end
			end
		end,

		Delete = function(self,profile)
			local selected = BossProfiles.profileList:GetSelected()

			-- Delete a non-root element
			if(profile.parent ~= nil) then
				local role = BossProfiles.role:GetSelected()

				-- Delete role profile
				if(profile.ref.type == "PROFILE" and profile.ref.data[role] ~= nil and profile.ref.data[role].activeSpec ~= 0) then
					profile.ref.data[role] = nil

				-- Delete entire profile
				else
					-- Find the target elements position
					local pos 
					for k,v in pairs(profile.parent.data) do
						if(v == profile.ref) then
							pos = k
							tremove(profile.parent.data,pos)
							break
						end
					end

					-- Handle selection changes if deleted element shares same parent as currently selected
					if(selected ~= nil) then

						-- If a common parent is shared
						if(selected.parent == profile.parent) then
							-- If only child was selected and deleted, select parent
							if(#profile.parent.data == 0) then
								local button
								for i=1,#BossProfilesFrame_ProfilesScrollFrame.buttons do
									button = BossProfilesFrame_ProfilesScrollFrame.buttons[i]
									if(button.data ~= nil and button.data.ref == profile.parent) then
										BossProfiles.profileList:SetSelected(button.data)
										break
									end
								end

							 -- If selected element was last child, select preceding sibling
							elseif(#profile.parent.data == pos-1) then
								local button
								for i=1,#BossProfilesFrame_ProfilesScrollFrame.buttons do
									button = BossProfilesFrame_ProfilesScrollFrame.buttons[i]
									if(button.data ~= nil and button.data.ref == profile.parent.data[pos-1]) then
										BossProfiles.profileList:SetSelected(button.data)
										break
									end
								end
							else
								BossProfiles.profileList:SetSelected(nil)
							end
						else
							BossProfiles.profileList:SetSelected(nil)
						end
					end
				end

			-- Assume root element and flush it
			else
				local found
				if(selected ~= nil) then
					for _,node in pairs(BossProfiles.profileList.cache) do
						-- Break and transfer selection to category if descendant is selected
						if(found and node == selected and selected.depth > profile.depth) then
							BossProfiles.profileList:SetSelected(profile)
							break
						end

						-- Break if sibling is found
						if(found and node.depth == profile.depth) then
							break
						end

						-- If category has been located, set to true
						if(node == profile) then found = true end
					end
				end

				-- Finally clear profile
				profile.ref.data = {}
			end

			-- Reselect the correct profile, this is necessary because the cache recycles tables
			-- instead of deleting them. If an element is deleted before the currently selected,
			-- the selected reference must be adjusted
			if(selected ~= nil) then
				local ref = selected.ref
				BossProfiles.profileList:RefreshData()
				for _,node in pairs(BossProfiles.profileList.cache) do
					if(node.ref == ref) then
						BossProfiles.profileList:SetSelected(node)
						break
					end
				end
			else
				BossProfiles.profileList:RefreshData()
			end

			-- Update profile list so that changes can be reflected
			BossProfiles.profileList:Update()	
		end,

		-- Gets subprofile data for current role/spec
		GetRelative = function()
			local selected = BossProfiles.profileList:GetSelected()

			if(selected and selected.ref.type == "PROFILE") then
				local role = BossProfiles.role:GetSelected()
				local spec = selected.ref.data[role].activeSpec
				return selected.ref.data[role][spec]
			else
				return nil
			end
		end,

		-- TODO: implement functionality for this
		Apply = function(self)
			local settings = BossProfiles.db.char.settings

			if(UnitAffectingCombat("player") == false) then
				local selected = BossProfiles.profileList:GetSelected()
				local role = BossProfiles.role:GetSelected()

				if(selected ~= nil) then
					if(selected.ref.type == "PROFILE" and selected.ref.data[role] ~= nil) then
						local profile = selected.ref.data[role]
						
						if(profile.activeSpec ~= 0) then
							if((settings.debug or settings.verbose) and self.lastLoaded ~= selected.ref.name) then 
								BossProfiles:Print("Applying "..selected.ref.name.." profile")
								self.lastLoaded = selected.ref.name
							end

			   			if(settings.spec and BossProfiles.content.spec:GetChangeCount(profile) > 0) then
								BossProfiles.content.spec:Apply(profile)
							elseif(settings.talents and BossProfiles.content.talents:GetChangeCount(profile) > 0) then
								BossProfiles.content.talents:Apply(profile)
							elseif(settings.glyphs and BossProfiles.content.glyphs:GetChangeCount(profile) > 0) then
								BossProfiles.content.glyphs:Apply(profile)
							elseif(settings.buffs and BossProfiles.content.buffs:GetChangeCount(profile) > 0) then
								BossProfiles.content.buffs:Apply(profile)
							elseif(settings.gear and BossProfiles.content.gear:GetChangeCount(profile) > 0) then
								BossProfiles.content.gear:Apply(profile)
							else
								BossProfiles:Print("No changes remaining")
							end

							-- Update the apply button each time it is clicked
							BossProfiles.util:UpdateApplyButton()
						else
							if(settings.debug or settings.verbose) then BossProfiles:Print(selected.ref.name.." does not have data for selected role") end
						end
					else
						if(settings.debug or settings.verbose) then BossProfiles:Print("You can't apply a category, please select a profile") end
					end
				else
					if(settings.debug or settings.verbose) then BossProfiles:Print("A profile selection is required") end
				end
			else
				if(settings.debug or settings.verbose) then BossProfiles:Print("Error - unable to load profile while in combat") end
			end
		end,
	}

	BossProfiles.role = {
		frame = BossProfilesFrame_ProfilesRoleButtons,
		background = BossProfilesFrame_ProfilesRoleButtonsBackground,
		selected = nil,
		rolesShown = 0,
		
		Update = function(self)
			self.rolesShown = 0
			local primaryRole, secondaryRole, button
			local primarySpec, secondarySpec = GetSpecialization(nil,nil,1), GetSpecialization(nil,nil,2)

			-- Get roles of primary and secondary specs
			if(primarySpec) then primaryRole = GetSpecializationRole(primarySpec) end
			if(secondarySpec) then secondaryRole = GetSpecializationRole(secondarySpec) end

			-- Set role icon positions, toggle visibility if playable
			for i=1,3 do
				button = self.frame.buttons[i]

				if((primaryRole and primaryRole == button.role) or (secondaryRole and secondaryRole == button.role)) then
					button:SetPoint("TOPRIGHT", (-35 * self.rolesShown + -4), -4)
					button:Show()
					self.rolesShown = self.rolesShown +1
				else
					button:Hide()
				end
			end

			-- Initialize role selection to role of currently active spec (if no button is currently selected)
			-- If active spec role is nil, use other specs role
			if(self.selected == nil) then 
				local spec = GetSpecialization()
				if(spec ~= nil) then 
					self:SetSelected(GetSpecializationRole(spec))
				elseif(primaryRole ~= nil) then
					self:SetSelected(primaryRole)
				elseif(secondaryRole ~= nil) then
					self:SetSelected(secondaryRole)
				end
			end
		end,

		GetSelected = function(self)
			return self.selected
		end,

		SetSelected = function(self,role)
			if(self.selected ~= role) then
				self.selected = role

				-- Set Background gradient to match active role
				if(role == "TANK") then
					self.background:SetVertexColor(0.2,0.2,0.9,0.5)
				elseif(role == "HEALER") then
					self.background:SetVertexColor(0.2,0.9,0.2,0.4)
				elseif(role == "DAMAGER") then
					self.background:SetVertexColor(0.9,0.2,0.2,0.5)
				end

				-- Desaturate inactive roles
				for i=1,3 do
					button = self.frame.buttons[i]

					if(button.role == role) then
						button:GetNormalTexture():SetDesaturated(false)
					else
						button:GetNormalTexture():SetDesaturated(true)
					end
				end

				-- Actions to perform after switching role
				BossProfiles.profileList:Update()

				if(BossProfiles.profileList:GetSelected() ~= nil) then
					BossProfiles.profile:Load()
				end

				BossProfiles.util:UpdateApplyButton()
			end
		end,
	}

	BossProfiles.profileList = {

		initialized, cache, selected,
		frame = BossProfilesFrame_ProfilesScrollFrame,

		Initialize = function(self)
			if(not self.initialized) then
				local width = ((1 - PROFILES_CONTENT_WIDTH) * BossProfilesFrame_Profiles:GetWidth()) - 23
				local height = BossProfilesFrame_Profiles:GetHeight() - 2

				self.frame:SetSize(width,height)

				local buttons = {}
				self.frame.buttons = buttons

				for i = 1, (ceil(height / PROFILES_LISTING_HEIGHT) + 1) do
					local button = CreateFrame("Button","BossProfilesFrame_ProfilesScrollFrameListing"..i,self.frame.scrollChild,"ProfileListingTemplate")

					button:SetSize(width-14,PROFILES_LISTING_HEIGHT)

					if i == 1 then
						button:SetPoint("TOPLEFT", 6, -1)
					else
						button:SetPoint("TOPLEFT", buttons[i - 1], "BOTTOMLEFT", 0, -PROFILES_LISTING_SPACING)
					end
					buttons[i] = button
				end

				self.frame.scrollBar.doNotHide = true
				self.frame.update = function() self:Update() end
				self.frame.dynamic = function(offset) local i=floor(offset/PROFILES_LISTING_HEIGHT) return i,offset-(i*PROFILES_LISTING_HEIGHT) end
				HybridScrollFrame_CreateButtons(self.frame, nil, nil, nil, nil, nil, nil, -PROFILES_LISTING_SPACING)

				self.initialized = true
			end
		end,

		Update = function(self)
			if(not self.cache) then self:RefreshData() end

			local index, button, listing
			local role = BossProfiles.role:GetSelected()
			-- local offset = 0

			for i = 1,#self.frame.buttons do
				button = self.frame.buttons[i]
				index = HybridScrollFrame_GetOffset(self.frame) + i -- + offset

				if index <= #self.cache then
					local data = self.cache[index]
					button.data = data

					-- if(data ~= nil and data.ref.collapsed) then
					-- 	local count = 0
					-- 	for i=index+count, #self.cache do
					-- 		if(self.cache[index+count+1].depth ~= data.depth) then
					-- 			count = count + 1
					-- 		else
					-- 			offset = offset + count
					-- 			break
					-- 		end
					-- 	end
					-- end

					if(data.ref.name ~= nil) then
						button.title:SetText(data.ref.name)
						button.title:SetCursorPosition(0)
					else
						button.title:SetText("Enter name...")
					end

					-- Enable text edit field if element has rename pending
					if(data.ref.renamePending) then
						button.title.background:SetAlpha(0.5)
						button.title:EnableMouse(true)
						button.title:Enable()
					else
						button.title.background:SetAlpha(0)
						button.title:EnableMouse(false)
						button.title:Disable()
					end

					-- If button is currently selected, set highlight to max opacity
					if(self.selected ~= nil and self.selected == data) then
						button.highlight:SetAlpha(1)
						button.highlight:Show()
					else
						button.highlight:Hide()
						button.highlight:SetAlpha(0.6)
					end

					if(data.ref.type == "PROFILE") then
						-- Hiding expand button because a profile can't have children
						button.expand:Hide()

						-- Set title size and offset
						button.title:SetPoint("LEFT",data.depth*PROFILES_LISTING_INDENT+5,0)
						button.title:SetWidth(button:GetWidth()-data.depth*PROFILES_LISTING_INDENT-12)

						-- If profile exists for current profile/spec combo or rename is pending, set text to max opacity
						if(data.ref.renamePending or (data.ref.data[role] ~= nil and data.ref.data[role].activeSpec ~= 0)) then
							button.title:SetAlpha(1)
						else
							button.title:SetAlpha(0.6)
						end

					elseif(data.ref.type == "CATEGORY") then
						-- Set expand/title size and offset
						button.expand:SetPoint("LEFT",data.depth*PROFILES_LISTING_INDENT,0)
						button.title:SetPoint("LEFT",(data.depth+1)*PROFILES_LISTING_INDENT+3,0)
						button.title:SetWidth(button:GetWidth()-(data.depth+1)*PROFILES_LISTING_INDENT-10)

						button.expand:Show()
						button.title:SetAlpha(1)

						-- Toggle expand button to Enabled state if node has children, otherwise disable
						if(data.hasChildren) then
							if(data.ref.collapsed) then
								button.expand:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
							else
								button.expand:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
							end
							button.expand:Enable()
						else
							button.expand:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Disabled")
							button.expand:Disable()
						end
					end

					button:Show()
				else
					button.data = nil
					button.title:SetText("")
					button:Hide()
				end
			end

			local listHeight = 
			HybridScrollFrame_Update(self.frame, 
				#self.cache * (PROFILES_LISTING_HEIGHT + PROFILES_LISTING_SPACING), 
				self.frame:GetHeight())
		end,

		GetSelected = function(self)
			return self.selected
		end,

		GetSelectedIndex = function(self)
			for k,v in pairs(self.frame.buttons) do
				if(self.selected == v.data) then
					return k
				end
			end
		end,

		SetSelected = function(self,node)
			self.selected = node
			BossProfiles.content:HideSpecArt()

			if(node ~= nil) then
				if(node.ref.type == "CATEGORY") then
					if(node.depth >= MAX_TREE_DEPTH) then
						BossProfilesFrame_ProfilesFooterNewCategoryButton:Disable()
					else
						BossProfilesFrame_ProfilesFooterNewCategoryButton:Enable()
					end

					if(node.depth > MAX_TREE_DEPTH) then
						BossProfilesFrame_ProfilesFooterNewProfileButton:Disable()
					else
						BossProfilesFrame_ProfilesFooterNewProfileButton:Enable()
					end
					BossProfiles.content.talents.frame:Hide()

				elseif(node.ref.type == "PROFILE") then
					if(node.depth >= MAX_TREE_DEPTH+1) then
						BossProfilesFrame_ProfilesFooterNewCategoryButton:Disable()
					else
						BossProfilesFrame_ProfilesFooterNewCategoryButton:Enable()
					end

					if(node.depth > MAX_TREE_DEPTH+1) then
						BossProfilesFrame_ProfilesFooterNewProfileButton:Disable()
					else
						BossProfilesFrame_ProfilesFooterNewProfileButton:Enable()
					end
					BossProfiles.profile:Load(node.ref)
					BossProfiles.content.talents.frame:Show()
				end
			else
				BossProfilesFrame_ProfilesFooterNewCategoryButton:Disable()
				BossProfilesFrame_ProfilesFooterNewProfileButton:Disable()
			end

			-- Finally update the frame
			self:Update()
			BossProfiles.util:UpdateApplyButton()
		end,

		ShiftSelected = function(self,pos)
			if(self.selected ~= nil and pos ~= nil and pos ~= 0) then
				for k,v in pairs(self.cache) do
					if(v.ref == self.selected.ref) then
						if self.cache[k+pos] ~= nil then
							self.selected = self.cache[k+pos]
							break
						end
					end
				end
			end
		end,

		-- Slower execution time compared to old function, however minimal garbage is produced
		RefreshData = function(self,table,parent,depth)
			-- On first run
			if table == nil then 
				table = BossProfiles.db.char.profiles
				depth = 0
				self.pos = 1
				if self.cache == nil then
					self.cache = {}
				end
			end

			for _,t in pairs(table) do
				local node = self.cache[self.pos]
				if(t.type == "CATEGORY") then
					if(node ~= nil) then
						node.ref = t
						node.parent = parent
						node.depth = depth
						node.hasChildren = #t.data>0
					else
						self.cache[self.pos]={ref=t,parent=parent,depth=depth,hasChildren=#t.data>0}
					end
					self.pos = self.pos+1

					if(not t.collapsed) then
						self:RefreshData(t.data,t,depth+1)
					end
				else
					if(node ~= nil) then
						node.ref=t
						node.parent=parent
						node.depth=depth
						node.hasChildren=false
					else
						self.cache[self.pos]={ref=t,parent=parent,depth=depth,hasChildren=false}
					end
					self.pos = self.pos+1
				end
			end

			-- If recycled table size is larger than new size, remove remaining elements
			-- Execess data will be garbage, eventually to be removed by garbage collector
			if(depth == 0 and #self.cache >= self.pos) then
				for i=self.pos,#self.cache do
					self.cache[i] = nil
				end
			end
		end,

		-- Faster execution time, but extremely inefficient. Generates a lot of garbage
		RefreshDataFast = function(self,table,parent,list,depth)
			if table == nil then table = BossProfiles.db.char.profiles end
			if list == nil then list = {} depth = 0 end

			for _,t in pairs(table) do
				if(t.type == "CATEGORY") then
					list[#list+1]={ref=t,parent=parent,depth=depth,hasChildren=#t.data>0}
					self:RefreshData(t.data,t,list,depth+1)
				else
					list[#list+1]={ref=t,parent=parent,depth=depth,hasChildren=false}
				end
			end

			self.cache = list
		end,

		CalcScroll = function(self,offset)
			local index = floor(offset/PROFILES_LISTING_HEIGHT)
			return index, offset - (index*PROFILES_LISTING_HEIGHT)
		end,

		-- Collapse or expand a category
		ToggleCategory = function(self,node)

			-- Collapse
			if(node.ref.collapsed == nil) then
				node.ref.collapsed = true

				local found
				for i=1,#self.cache do
					-- Break if selected element is found and transfer selection to category
					if(found and self.cache[i] == self.selected and self.selected.depth > node.depth) then
						self:SetSelected(node)
						break
					end

					-- Break if sibling is found
					if(found and self.cache[i].depth == node.depth) then
						break
					end

					-- If category has been located, set to true
					if(self.cache[i] == node) then found = true end
				end

			-- Expand
			elseif(node.ref.collapsed) then
				node.ref.collapsed = nil
			end

			-- reselect correct element, to account for cache recycling
			if(self.selected ~= nil) then
				local ref = self.selected.ref
				BossProfiles.profileList:RefreshData()
				for i=1,#self.cache do
					if(self.cache[i].ref == ref) then
						self:SetSelected(self.cache[i])
						break
					end
				end
			else
				BossProfiles.profileList:RefreshData()
			end

			BossProfiles.profileList:Update()
		end,
	}

	BossProfiles.content = {
		initialized,
		frame = BossProfilesFrame_ProfilesContentFrame,
		specArt = BossProfilesFrame_ProfilesContentFrameSpecArt,

		Initialize = function(self)
			if(not self.initialized) then 
				-- Initialize content frame size using relative values
				local profilesFrame = BossProfilesFrame_Profiles
				self.frame:SetSize(PROFILES_CONTENT_WIDTH * profilesFrame:GetWidth(), profilesFrame:GetHeight() - 2)


				self.intialized = true
			end
		end,

		ShowSpecArt = function(self)
			local role = BossProfiles.role:GetSelected()
			local profile = BossProfiles.profileList:GetSelected()
			if(profile and profile.ref.data[role].activeSpec ~= 0) then
				self.specArt:SetTexture("Interface\\TALENTFRAME\\"..(select(5, GetSpecializationInfo(profile.ref.data[role].activeSpec))))
				self.specArt:Show()
			end
		end,

		HideSpecArt = function(self)
			self.specArt:Hide()
		end,

		spec = {
			initialized,
			frame = BossProfilesFrame_ProfilesContentFrameTalentsSpecDropDown,

			Initialize = function(self)
				UIDropDownMenu_Initialize(self.frame, function() self:Update() end)
				UIDropDownMenu_SetText(self.frame, "Select Spec")
				UIDropDownMenu_SetWidth(self.frame, 150, 0)
				UIDropDownMenu_SetButtonWidth(self.frame, 250)
			end,

			Update = function(self)
				local info, specName, specIcon
				local primarySpec = GetSpecialization(nil,nil,1)
				local secondarySpec = GetSpecialization(nil,nil,2)

				for i=1, GetNumSpecializations() do
					if(GetSpecializationRole(i) == BossProfiles.role:GetSelected()) then
						_,specName,_,specIcon = GetSpecializationInfo(i)

						info = UIDropDownMenu_CreateInfo()
						info.func = 
							function(self, arg1, arg2, alreadySelected)
								if (not alreadySelected) then
									BossProfiles.content.spec:SetSelected(self.value, false, "RELATIVE")
									UIDropDownMenu_SetSelectedValue(UIDROPDOWNMENU_OPEN_MENU, self.value)
								end
							end
						info.icon = specIcon

						-- Set relative spec, 0: not specced, 1: primary, 2: secondary
						if(i == primarySpec) then 
							info.value = 1
							info.text = specName.." (Primary)"
						elseif(i == secondarySpec) then 
							info.value = 2
							info.text = specName.." (Secondary)"
						else 
							info.value = 0
							info.text = specName.." (Inactive)"
							info.disabled = true
						end

						-- Finally add the button
						UIDropDownMenu_AddButton(info)
					end
				end

				-- Clear spec selection if active spec is currently not playable
				local profile = BossProfiles.profileList:GetSelected()
				if(profile ~= nil) then
					local role = BossProfiles.role:GetSelected()
					local spec = profile.ref.data[role].activeSpec
					if(spec ~= primarySpec and spec ~= secondarySpec) then
						self:SetSelected(0)
						profile.ref.data[role].activeSpec = 0
					end
				end
			end,

			GetSelected = function(self)
				return self.frame.value
			end,

			SetSelected = function(self,spec,force,type)
				-- if no spec has been provided, clear current selection
				if(spec == nil or spec == 0) then
					self.frame.selectedValue = nil
					_G[self.frame:GetName().."Text"]:SetText("Select Spec")
					BossProfiles.content:HideSpecArt()

				-- Else if a different spec has been clicked on
				else
					local relative, real
					if(type=="REAL") then
						relative = BossProfiles.util:GetRelativeSpec(spec)
						real = spec
					elseif(type=="RELATIVE") then
						relative = spec
						real = GetSpecialization(nil,nil,spec)
					end

					if(relative ~= self:GetSelected()) then
						local role = BossProfiles.role:GetSelected()
						local profile = BossProfiles.profileList:GetSelected()

						-- Set the activeSpec to new spec
						profile.ref.data[role].activeSpec = real

						-- force selection if option is being set indirectly (not via menu item click)
						if(force and relative ~= 0) then
							-- Set the internal value
							UIDropDownMenu_SetSelectedValue(self.frame, relative)

							-- Force selected text to be spec name
							local text = select(2,GetSpecializationInfo(real))
							if(relative == 1) then
								text = text.." (Primary)"
							else
								text = text.." (Secondary)"
							end

							_G[self.frame:GetName().."Text"]:SetText(text)
						end

						-- Actions to perform after changing spec
						BossProfiles.content:ShowSpecArt()
					end
				end

				-- Actions to perform
				BossProfiles.profileList:Update()
				BossProfiles.content.talents:Update()
				BossProfiles.util:UpdateApplyButton()
				BossProfilesFrame_ProfilesContentFrameTalents:Show()
			end,

			GetChangeCount = function(self,profile)
				if(profile ~= nil) then
					if(profile.activeSpec == 0 or profile.activeSpec == GetSpecialization()) then
						return 0
					else
						local relSpec = BossProfiles.util:GetRelativeSpec(profile.activeSpec) 
						if(relSpec ~= 0 and relSpec ~= GetActiveSpecGroup()) then
							return 1
						else
							return -1
						end
					end
				else
					return -1
				end
			end,

			Apply = function(self,profile)
				local settings = BossProfiles.db.char.settings
				if(self:GetChangeCount(profile) > 0) then
					local relSpec = BossProfiles.util:GetRelativeSpec(profile.activeSpec) 
					local specName = select(2,GetSpecializationInfo(profile.activeSpec))
					local castName,_,_,_,_,endTime = UnitCastingInfo("Player")

					if(settings.debug or settings.verbose) then 
						if(castName ~= nil and castName == "Activating Primary Spec" or castName == "Activating Secondary Spec") then
							BossProfiles:Print("Currently switching spec, please wait "..floor(endTime/1000-GetTime()).."s")
						else
							if(relSpec == 1) then
								BossProfiles:Print("Switching to "..specName.." (Primary)") 
							elseif(relSpec == 2) then
								BossProfiles:Print("Switching to "..specName.." (Secondary)") 
							end
						end
					end

					SetActiveSpecGroup(relSpec)
				end
			end,
		},

		talents = {
			initialized, activeBackdrop, inactiveBackdrop,
			frame = BossProfilesFrame_ProfilesContentFrameTalents,

			Initialize = function(self)
				if(not self.initialized) then

					self.activeBackdrop = {
					   bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
					   insets = {left = 2, right = 2, top = 2, bottom = 2},
					   edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
					   edgeSize = 12,
					}

					self.inactiveBackdrop = {
					   bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
					   insets = {left = 2, right = 2, top = 2, bottom = 2},
					}

					-- Talent Buttons
					local button
					local buttons = {}
					self.frame.buttons = buttons
					for tier=1,7 do
						for col=1,3 do
							-- Create the button
							button = CreateFrame("Button","$parentTalentRow"..tier.."Column"..col,self.frame,"ProfileTalentButtonTemplate")

							-- Add button to array
							tinsert(self.frame.buttons, button)

							-- Set button positioning
							if(tier==1 and col==1) then
								button:SetPoint("TOPLEFT", 9, -78)
							elseif(col == 1) then
								button:SetPoint("TOPLEFT", buttons[#buttons-3], "BOTTOMLEFT", 0, -PROFILES_TALENT_SPACING)
							else
								button:SetPoint("LEFT",buttons[#buttons-1],"RIGHT",PROFILES_TALENT_SPACING,0)
							end

							-- Format button
							button:SetSize(PROFILES_TALENT_WIDTH, PROFILES_TALENT_HEIGHT)
							button.icon:SetSize(PROFILES_TALENT_HEIGHT-10, PROFILES_TALENT_HEIGHT-10)
							button.tier = tier
							button.col = col
						end
					end

					self.initialized = true
				end
			end,

			Update = function(self)
				local profile = BossProfiles.profileList:GetSelected()

				if(profile ~= nil and profile.ref.type == "PROFILE") then
					local button, id, name, texture, active, selected
					local maxTier = GetMaxTalentTier()
					local role = BossProfiles.role:GetSelected()
					local realSpec = profile.ref.data[role].activeSpec
					local relativeSpec = BossProfiles.util:GetRelativeSpec(realSpec)
					local relativeProfile = BossProfiles.profile:GetRelative()

					-- Initialize talent profile if none already exists
					if(relativeProfile and not relativeProfile.talents) then relativeProfile.talents = {0,0,0,0,0,0,0} end

					
					for i=1, #self.frame.buttons do
						button = self.frame.buttons[i]

						-- If a relative spec exists, load talents for that spec
						if(relativeSpec ~= 0) then
							id, title, texture, active = GetTalentInfo(button.tier,button.col,relativeSpec)

							-- Set frame border for actual selected talents
							if(active) then
								button:SetBackdrop(self.activeBackdrop)
								button:SetBackdropColor(0,0,0,0.3)
							else
								button:SetBackdrop(self.inactiveBackdrop)
								button:SetBackdropColor(0,0,0,0.3)
							end

							-- Check to see if talent is selected in profile sandbox
							if(button.col == relativeProfile.talents[button.tier]) then
								selected = true
							else
								selected = false
							end

							-- Set attributes
							button.title:SetText(title)
							button.icon:SetTexture(texture)
							button.talentID = id

							-- Style talent buttons
							if(button.tier <= maxTier and selected) then
								button.icon:SetDesaturated(false)
								button.icon:SetAlpha(1)
								button.title:SetAlpha(1)	
							elseif(button.tier <= maxTier and not selected) then
								button.icon:SetDesaturated(true)
								button.icon:SetAlpha(0.8)
								button.title:SetAlpha(0.8)
							elseif(button.tier > maxTier) then
								button.icon:SetDesaturated(true)
								button.icon:SetAlpha(0.2)
								button.title:SetAlpha(0.2)
							end

							-- if(button.tier <= maxTier and selected) then
							-- 	button.icon:SetDesaturated(false)
							-- 	button.icon:SetAlpha(1)
							-- 	button.icon:SetVertexColor(1,1,1)
							-- 	button.title:SetAlpha(1)	
							-- elseif(button.tier <= maxTier and not selected) then
							-- 	button.title:SetAlpha(0.8)
							-- 	if(active) then
							-- 		button.icon:SetDesaturated(false)
							-- 		button.icon:SetAlpha(0.7)
							-- 		button.icon:SetVertexColor(0.4,0.4,0.4)
							-- 	else
							-- 		button.icon:SetDesaturated(true)
							-- 		button.icon:SetAlpha(0.8)
							-- 	end
							-- elseif(button.tier > maxTier) then
							-- 	button.icon:SetDesaturated(true)
							-- 	button.icon:SetAlpha(0.2)
							-- 	button.title:SetAlpha(0.2)
							-- 	button.icon:SetVertexColor(1,1,1)
							-- else
							-- 	button.icon:SetVertexColor(1,1,1)
							-- end
						-- Else load default name/textures
						else
							button:SetBackdrop(self.inactiveBackdrop)
							button:SetBackdropColor(0,0,0,0.3)
							button.title:SetText("Talent Name")
							button.title:SetAlpha(0.2)
							button.icon:SetTexture("Interface\\Icons\\INV_MISC_QUESTIONMARK")
							button.icon:SetAlpha(0.2)
							button.icon:SetDesaturated(true)
						end
					end
				end
			end,

			GetChangeCount = function(self,profile)
				if(profile ~= nil) then
					local relSpec = BossProfiles.util:GetRelativeSpec(profile.activeSpec)

					if(relSpec ~= 0) then
						local count = 0
						local talent, selected, available

						for tier=1,GetMaxTalentTier() do
							for col=1,3 do
								talent = profile[profile.activeSpec].talents[tier]
								selected, available = select(4,GetTalentInfo(tier,col,relSpec))

								if(talent ~= nil and talent ~= 0 and (
									(not selected and available and col == talent) or (selected and col ~= talent)
								)) then
									count = count + 1
								end
							end
						end

						return count
					else
						return -1
					end
				else
					return -1
				end
			end,
			
			Apply = function(self,profile)
				if(profile ~= nil) then
					local settings = BossProfiles.db.char.settings
					local relSpec = BossProfiles.util:GetRelativeSpec(profile.activeSpec)
					if(relSpec == GetActiveSpecGroup()) then
						local id,selected,available
						for tier,col in pairs(profile[profile.activeSpec].talents) do
							id,_,_,selected,available = GetTalentInfo(tier,col,relSpec)
							if(col ~= 0 and not selected) then
								BossProfilesLoader:SetAttribute("type","macro")
								BossProfilesLoader:SetAttribute("macrotext",
									"/stopmacro [combat]\n"..
									"/click TalentMicroButton\n"..
									"/click TalentMicroButton\n"..
									"/click PlayerSpecTab"..relSpec.."\n"..
									"/click PlayerTalentFrameTalentsTalentRow"..tier.."Talent"..col.."\n"..
									"/click StaticPopup1Button1"
								)

								C_Timer.After(0.1, function() LearnTalents(id) end)

								if(settings.debug) then BossProfiles:Print("Learning tier "..tier..", column "..col) end
								return false
							else
								if(settings.debug) then BossProfiles:Print("No talent for tier "..tier.." needs to be removed") end
								BossProfiles.util:PreventModification()
							end
						end
					end
				end
			end,
		},
	}
end